@@ -28,7 +28,9 @@ oneTimeSetUp() {
 
 	: ${GIT_FTP_USER=ftp}
 	: ${GIT_FTP_PASSWD=}
-	: ${GIT_FTP_ROOT=localhost}
+	: ${GIT_FTP_HOST=localhost}
+	: ${GIT_FTP_PORT=:21}
+	: ${GIT_FTP_ROOT=}
 
 	START=$(date +%s)
 }
@@ -48,9 +50,13 @@ setUp() {
 	fi
 	GIT_PROJECT_NAME="$(basename $GIT_PROJECT_PATH)"
 
-	GIT_FTP_URL="$GIT_FTP_ROOT/$GIT_PROJECT_NAME"
+	REMOTE_BASE_URL="ftp://$GIT_FTP_HOST$GIT_FTP_PORT"
+	REMOTE_PATH="$GIT_FTP_ROOT/$GIT_PROJECT_NAME"
+	GIT_FTP_URL="$REMOTE_BASE_URL/$REMOTE_PATH"
 
-	CURL_URL="ftp://$GIT_FTP_USER:$GIT_FTP_PASSWD@$GIT_FTP_URL"
+	[ -n "$GIT_FTP_USER" ] && CURL_ARGS=" -u $GIT_FTP_USER:$GIT_FTP_PASSWD"
+	CURL="curl $CURL_ARGS"
+	CURL_URL="$GIT_FTP_URL"
 
 	[ -n "$GIT_FTP_USER" ] && GIT_FTP_USER_ARG="-u $GIT_FTP_USER"
 	[ -n "$GIT_FTP_PASSWD" ] && GIT_FTP_PASSWD_ARG="-p $GIT_FTP_PASSWD"
@@ -77,18 +83,18 @@ tearDown() {
 	assertEquals '' "$tmpfiles"
 	rm -rf $GIT_PROJECT_PATH
 	command -v lftp >/dev/null 2>&1 && {
-		lftp -u $GIT_FTP_USER,$GIT_FTP_PASSWD $GIT_FTP_ROOT -e "set ftp:list-options -a; rm -rf '$GIT_PROJECT_NAME'; exit" > /dev/null 2>&1
+		lftp -u "$GIT_FTP_USER,$GIT_FTP_PASSWD" "$GIT_FTP_URL" -e "set ftp:list-options -a; rm -rf '$GIT_PROJECT_NAME'; exit" > /dev/null 2>&1
 	}
 }
 
 test_displays_usage() {
 	usage=$($GIT_FTP_CMD 2>&1)
-	assertEquals "git-ftp <action> [<options>] <url>" "$usage"
+	assertEquals "git-ftp <action> [<options>] [<url>]" "$usage"
 }
 
 test_prints_version() {
 	version=$($GIT_FTP_CMD 2>&1 --version)
-	assertEquals = "git-ftp version 1.2.0"  "$version"
+	assertEquals "git-ftp version 1.3.1"  "$version"
 }
 
 test_inits() {
@@ -101,8 +107,9 @@ test_inits() {
 test_init_fails() {
 	init=$($GIT_FTP_CMD -v -u wrong_user -p wrong_passwd $GIT_FTP_URL init 2>&1)
 	assertEquals 5 $?
-	error_count=$(echo "$init" | grep -F 'Access denied' | wc -l)
-	assertEquals 1 $error_count
+	# Not all servers respond correctly
+	#error_count=$(echo "$init" | grep -F 'Access denied' | wc -l)
+	#assertEquals 1 $error_count
 }
 
 test_inits_and_pushes() {
@@ -139,7 +146,7 @@ test_pushes_and_fails() {
 
 test_push_unknown_commit() {
 	$GIT_FTP init > /dev/null
-	echo '000000000' | curl -s -T - $CURL_URL/.git-ftp.log
+	echo '000000000' | $CURL -s -T - $CURL_URL/.git-ftp.log
 	push="$($GIT_FTP push 0>&- 2>&1)"
 	assertEquals 0 $?
 	assertContains 'Unknown SHA1 object' "$push"
@@ -171,7 +178,7 @@ test_push_added() {
 	git commit -m "change" > /dev/null 2>&1
 	push=$($GIT_FTP push)
 	assertEquals 0 $? || echo "Push: $push"
-	assertEquals "1" "$(curl -s $CURL_URL/$file)"
+	assertEquals "1" "$($CURL -s $CURL_URL/$file)"
 }
 
 test_push_twice() {
@@ -194,7 +201,7 @@ test_push_unknown_sha1() {
 	echo "1" >> "./test 1.txt"
 	git commit -a -m "change" > /dev/null 2>&1
 	# change remote SHA1
-	echo '000000000' | curl -T - $CURL_URL/.git-ftp.log 2> /dev/null
+	echo '000000000' | $CURL -T - $CURL_URL/.git-ftp.log 2> /dev/null
 	push=$(echo 'N' | $GIT_FTP push)
 	assertEquals 0 $?
 	echo "$push" | grep 'Unknown SHA1 object' > /dev/null
@@ -208,7 +215,7 @@ test_push_unknown_sha1_Y() {
 	echo "1" >> "./test 1.txt"
 	git commit -a -m "change" > /dev/null 2>&1
 	# change remote SHA1
-	echo '000000000' | curl -T - $CURL_URL/.git-ftp.log 2> /dev/null
+	echo '000000000' | $CURL -T - $CURL_URL/.git-ftp.log 2> /dev/null
 	push=$(echo 'Y' | $GIT_FTP push)
 	assertEquals 0 $?
 	echo "$push" | grep 'Unknown SHA1 object' > /dev/null
@@ -384,7 +391,7 @@ test_ignore_single_file_force_unknown_commit() {
 	echo $file > .git-ftp-ignore
 	git add .
 	git commit -m 'added new file that should be ignored' -q
-	echo '000000000' | curl -s -T - $CURL_URL/.git-ftp.log
+	echo '000000000' | $CURL -s -T - $CURL_URL/.git-ftp.log
 	push=$($GIT_FTP push -f)
 	assertFalse 'test failed: file was not ignored' "remote_file_exists '$file'"
 }
@@ -440,11 +447,9 @@ test_ignore_wildcard_files() {
 }
 
 test_include_init() {
-	cd $GIT_PROJECT_PATH
 	echo 'unversioned' > unversioned.txt
 	echo 'unversioned.txt' >> .gitignore
 	echo 'unversioned.txt:test 1.txt' > .git-ftp-include
-	echo 'new content' >> 'test 1.txt'
 	git add .
 	git commit -m 'unversioned file unversioned.txt should be uploaded with test 1.txt' > /dev/null
 	init=$($GIT_FTP init)
@@ -581,6 +586,31 @@ test_include_ftp_ignore_push() {
 	assertFalse ' .htaccess.prod was uploaded' "remote_file_exists '.htaccess.prod'"
 }
 
+test_include_always_syncroot_push() {
+	init=$($GIT_FTP init)
+	mkdir "public_html"
+	echo "public_html" > .gitignore
+	echo "content" > "public_html/always.html"
+	echo "!public_html/always.html" > .git-ftp-include
+	git add .
+	git commit -m "setup" > /dev/null
+	push="$($GIT_FTP push --syncroot "public_html")"
+	assertTrue " upload always.html" "remote_file_exists 'always.html'"
+}
+
+test_include_syncroot_push() {
+	init=$($GIT_FTP init)
+	mkdir "public_html"
+	echo "content" > "public_html/style.sass"
+	echo "content" > "public_html/style.css"
+	echo "public_html/style.css" > .gitignore
+	echo "public_html/style.css:style.sass" > .git-ftp-include
+	git add .
+	git commit -m "setup" > /dev/null
+	push="$($GIT_FTP push --syncroot "public_html")"
+	assertTrue " upload style.css" "remote_file_exists 'style.css'"
+}
+
 # addresses issue #41
 test_include_similar() {
 	cd $GIT_PROJECT_PATH
@@ -596,6 +626,50 @@ test_include_similar() {
 	assertTrue 'templates/foo.html was not uploaded' "remote_file_exists 'templates/foo.html'"
 }
 
+# resolves issue #245
+test_include_syncroot() {
+	source1="main.scss"
+	source2="other.scss"
+	source3="three.scss"
+	targetfile1="main.css"
+	targetfile2="other.css"
+	targetfile3="three.css"
+	syncroot="dist"
+	target1="$syncroot/$targetfile1"
+	target2="$syncroot/$targetfile2"
+	target3="$syncroot/$targetfile3"
+	touch "$source1" "$source2" "$source3"
+	mkdir "$syncroot"
+	touch "$target1" "$target2" "$target3"
+	echo "$syncroot" > ".gitignore"
+	echo "$target1:/$source1" >> ".git-ftp-include"
+	git add .
+	git commit -a -m "no dist files included" -q
+	init="$($GIT_FTP init --syncroot "$syncroot")"
+	assertTrue "remote file '$targetfile1'" "remote_file_exists '$targetfile1'"
+	assertFalse "remote file '$targetfile2'" "remote_file_exists '$targetfile2'"
+	assertFalse "remote file '$targetfile3'" "remote_file_exists '$targetfile3'"
+
+	# push no files
+	echo "$target2:/$source2" >> ".git-ftp-include"
+	echo "$target3:/$source3" >> ".git-ftp-include"
+	git add .
+	git commit -a -m "test not uploading dist files" -q
+	push="$($GIT_FTP push --syncroot "$syncroot")"
+	assertTrue "remote file '$targetfile1'" "remote_file_exists '$targetfile1'"
+	assertFalse "remote file '$targetfile2'" "remote_file_exists '$targetfile2'"
+	assertFalse "remote file '$targetfile3'" "remote_file_exists '$targetfile3'"
+
+	# push no files
+	echo "content" >> "$source2"
+	git add .
+	git commit -a -m "test uploading some dist files" -q
+	push="$($GIT_FTP push --syncroot "$syncroot")"
+	assertTrue "remote file '$targetfile1'" "remote_file_exists '$targetfile1'"
+	assertTrue "remote file '$targetfile2'" "remote_file_exists '$targetfile2'"
+	assertFalse "remote file '$targetfile3'" "remote_file_exists '$targetfile3'"
+}
+
 test_hidden_file_only() {
 	cd $GIT_PROJECT_PATH
 	echo "test" > .htaccess
@@ -630,6 +704,8 @@ test_file_with_nonchar() {
 # issue #209
 test_file_with_unicode() {
 	supports_unicode || startSkipping
+	# TODO: Get this working on MacOS if possible with curl
+	[ "$TRAVIS_OS_NAME" = "osx" ] && startSkipping
 	file1='umlaut_ä.md'
 	file1enc='umlaut_%C3%A4.md'
 	echo 'content' > "$file1"
@@ -643,6 +719,26 @@ test_file_with_unicode() {
 	assertFalse "file $file1 still exists in $CURL_URL" "remote_file_exists '$file1enc'"
 }
 
+# issue #259
+test_file_with_dash() {
+	dir='-dashdir'
+	file='-dash'
+	mkdir -- "$dir"
+	echo 'content' > "$file"
+	echo 'content' > "$dir/$file"
+	git add .
+	git commit -a -m 'added special filenames' -q
+	init="$($GIT_FTP init)"
+	assertTrue " file $file not uploaded" "remote_file_equals '$file'"
+	assertTrue " file $dir/$file not uploaded" "remote_file_equals '$dir/$file'"
+	git rm -q -- "$file"
+	git rm -q -- "$dir/$file"
+	git commit -m 'delete' -q
+	push="$($GIT_FTP push)"
+	assertFalse "file $file still exists in $CURL_URL" "remote_file_exists '$file'"
+	assertFalse "file $dir/$file still exists in $CURL_URL" "remote_file_exists '$dir/$file'"
+}
+
 test_syncroot() {
 	cd $GIT_PROJECT_PATH
 	syncroot='foo bar'
@@ -658,7 +754,7 @@ test_download() {
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
 	echo 'foreign content' > external.txt
-	curl -T external.txt $CURL_URL/ 2> /dev/null
+	$CURL -T external.txt $CURL_URL/ 2> /dev/null
 	rtrn=$?
 	assertEquals 0 $rtrn
 	rm external.txt
@@ -672,7 +768,7 @@ test_download_untracked() {
 	skip_without lftp
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
-	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	echo 'foreign content' | $CURL -T - $CURL_URL/external.txt 2> /dev/null
 	touch 'untracked.file'
 	$GIT_FTP download > /dev/null 2>&1
 	assertEquals 8 $?
@@ -688,7 +784,7 @@ test_download_syncroot() {
 	git commit -a -m "syncroot test" > /dev/null 2>&1
 	init=$($GIT_FTP init --syncroot foobar)
 	echo 'foreign content' > external.txt
-	curl -T external.txt $CURL_URL/ 2> /dev/null
+	$CURL -T external.txt $CURL_URL/ 2> /dev/null
 	rm external.txt
 	$GIT_FTP download --syncroot foobar/ > /dev/null 2>&1
 	rtrn=$?
@@ -701,7 +797,7 @@ test_download_dry_run() {
 	skip_without lftp
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
-	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	echo 'foreign content' | $CURL -T - $CURL_URL/external.txt 2> /dev/null
 	$GIT_FTP download --dry-run > /dev/null 2>&1
 	assertEquals 0 $?
 	assertTrue ' external file downloaded' "[ ! -e 'external.txt' ]"
@@ -712,7 +808,7 @@ test_pull() {
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
 	echo 'foreign content' > external.txt
-	curl -T external.txt $CURL_URL/ 2> /dev/null
+	$CURL -T external.txt $CURL_URL/ 2> /dev/null
 	rm external.txt
 	echo 'own content' > internal.txt
 	git add . > /dev/null 2>&1
@@ -737,7 +833,7 @@ test_pull_branch() {
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
 	echo 'foreign content' > external.txt
-	curl -T external.txt $CURL_URL/ 2> /dev/null
+	$CURL -T external.txt $CURL_URL/ 2> /dev/null
 	rm external.txt
 	echo 'own content' > internal.txt
 	git add . > /dev/null 2>&1
@@ -759,7 +855,7 @@ test_pull_no_commit() {
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
 	echo 'foreign content' > external.txt
-	curl -T external.txt $CURL_URL/ 2> /dev/null
+	$CURL -T external.txt $CURL_URL/ 2> /dev/null
 	rm external.txt
 	echo 'own content' > internal.txt
 	git add . > /dev/null 2>&1
@@ -776,7 +872,7 @@ test_pull_dry_run() {
 	skip_without lftp
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
-	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	echo 'foreign content' | $CURL -T - $CURL_URL/external.txt 2> /dev/null
 	echo 'own content' > internal.txt
 	git add . > /dev/null 2>&1
 	git commit -a -m "local modification" > /dev/null 2>&1
@@ -791,7 +887,7 @@ test_pull_untracked() {
 	skip_without lftp
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
-	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	echo 'foreign content' | $CURL -T - $CURL_URL/external.txt 2> /dev/null
 	echo 'own content' > internal.txt
 	echo 'internal.txt' >> .gitignore
 	git add . > /dev/null 2>&1
@@ -806,7 +902,7 @@ test_pull_stash() {
 	skip_without lftp
 	cd $GIT_PROJECT_PATH
 	$GIT_FTP init > /dev/null
-	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	echo 'foreign content' | $CURL -T - $CURL_URL/external.txt 2> /dev/null
 	echo 'own content' > internal.txt
 	git stash -u -q
 	pull=$($GIT_FTP pull 2> /dev/null)
@@ -818,6 +914,49 @@ test_pull_stash() {
 	assertEquals '' "$(git log -- 'internal.txt')"
 }
 
+test_pull_changedonly() {
+	skip_without lftp
+	echo 'foreign content' > external.txt
+	curl -T external.txt "$CURL_URL/" 2> /dev/null
+	rm external.txt
+	echo 'own content' > not_modified.txt
+	echo 'own content' > locally_modified.txt
+	echo 'own content' > locallyremotely_modified.txt
+	git add . > /dev/null 2>&1
+	git commit -a -m "init" -q
+	$GIT_FTP init -n
+	echo 'own content local modification' > locally_modified.txt
+	echo 'own content\nforeign content' > locallyremotely_modified.txt
+	$CURL -s -T "locallyremotely_modified.txt" "$CURL_URL/"
+	echo 'local modification\nown content' > locallyremotely_modified.txt
+	git add . > /dev/null 2>&1
+	git commit -a -m "local modification" -q
+	$GIT_FTP pull --changed-only > /dev/null 2>&1
+	assertEquals 0 $?
+	assertFalse ' external file downloaded' "[ -r 'external.txt' ]"
+	assertFalse ' external file listed in commit message' "git log | grep 'external.txt'"
+	assertFalse ' not_modified file listed in commit message' "git log | grep 'not_modified.txt'"
+	assertFalse ' locally_modified file listed in commit message' "git log | grep 'locally_modified.txt'"
+	assertTrue ' locallyremotely_modified file not listed in commit message' "git log | grep 'locallyremotely_modified.txt'"
+	lines="$(wc -l < locallyremotely_modified.txt | tr -d ' ')"
+	assertEquals ' locallremotely_modified not correctly merged' 3 "$lines"
+}
+
+test_snapshot_fail() {
+	skip_without lftp
+	$GIT_FTP init -n
+	$GIT_FTP snapshot -n
+	assertEquals 2 $?
+}
+
+test_snapshot() {
+	skip_without lftp
+	$GIT_FTP init -n
+	$CURL -s -Q "-*DELE $REMOTE_PATH/.git-ftp.log" "$REMOTE_BASE_URL" > /dev/null
+	$GIT_FTP snapshot -n
+	assertEquals 0 $?
+}
+
 test_submodule() {
 	submodule='sub'
 	file='file.txt'
@@ -834,6 +973,26 @@ test_submodule() {
 	assertTrue "test failed: $file not there as expected" "remote_file_exists '$submodule/$file'"
 }
 
+test_submodule_netrc() {
+	submodule='sub'
+	file='file.txt'
+	mkdir "$submodule"
+	cd "$submodule"
+	touch "$file"
+	git init -q
+	git add .
+	git commit -m 'initial submodule commit' -q
+	cd ..
+	git submodule -q add "/$submodule" > /dev/null
+	echo "machine localhost login $GIT_FTP_USER password $GIT_FTP_PASSWD" > .netrc
+	chmod 600 .netrc
+	git add .
+	git commit -m 'adding submodule' -q
+	# Setting $HOME to look for .netrc file
+	init="$(HOME="$(pwd)" $GIT_FTP_CMD init "$GIT_FTP_URL")"
+	assertTrue "test failed: $file not there as expected" "remote_file_exists '$submodule/$file'"
+}
+
 test_submodule_catchup() {
 	submodule='sub'
 	file='file.txt'
@@ -850,6 +1009,69 @@ test_submodule_catchup() {
 	assertTrue "test failed: $submodule/.git-ftp.log not there as expected" "remote_file_exists '$submodule/.git-ftp.log'"
 }
 
+test_submodule_syncroot() {
+	syncroot='dist'
+	submodule="sub"
+	submodule_path="$syncroot/sub"
+	file='file.txt'
+	mkdir -p "$submodule_path"
+	cd "$submodule_path"
+	touch "$file"
+	git init -q
+	git add .
+	git commit -m 'initial submodule commit' -q
+	cd - > /dev/null
+	git submodule -q add "./$submodule_path" "$submodule_path" > /dev/null
+	git commit -m 'adding submodule' -q
+	init="$($GIT_FTP init --syncroot "$syncroot")"
+	assertTrue "test failed: $file not there as expected" "remote_file_exists '$submodule/$file'"
+}
+
+test_pre_push() {
+	# init
+	hook=".git/hooks/pre-ftp-push"
+	message="pre-ftp-push executed."
+	echo "echo '$message'" > "$hook"
+	chmod +x "$hook"
+	out="$($GIT_FTP init)"
+	firstline="$(echo "$out" | head -n 1)"
+	assertEquals "$message" "$firstline"
+
+	# push
+	echo 'new content' >> 'test 1.txt'
+	git rm 'test 2.txt' -q
+	git commit -a -m 'new content' -q
+	echo 'while read -r -d "" status file; do echo "$status - $file"; done' > "$hook"
+	out="$($GIT_FTP push -n)"
+	firstline="$(echo "$out" | head -n 1)"
+	lastline="$(echo "$out" | tail -n 1)"
+	assertEquals "A - test 1.txt" "$firstline"
+	assertEquals "D - test 2.txt" "$lastline"
+
+	# push fail
+	echo 'new content' >> 'test 1.txt'
+	git commit -a -m 'new content' -q
+	echo 'exit 1' > "$hook"
+	out="$($GIT_FTP push -n)"
+	assertEquals 9 "$?"
+	assertEquals "" "$out"
+
+	# ignore hook
+	out="$($GIT_FTP push -n --no-verify)"
+	assertEquals 0 "$?"
+	assertEquals "" "$out"
+}
+
+test_post_push() {
+	# init
+	hook=".git/hooks/post-ftp-push"
+	message="post-ftp-push executed."
+	echo "echo '$message'" > "$hook"
+	chmod +x "$hook"
+	out="$($GIT_FTP init -n)"
+	assertEquals "$message" "$out"
+}
+
 disabled_test_file_named_dash() {
 	cd $GIT_PROJECT_PATH
 	echo "foobar" > -
@@ -862,14 +1084,14 @@ disabled_test_file_named_dash() {
 }
 
 remote_file_exists() {
-	curl "$CURL_URL/$1" --head > /dev/null
+	$CURL "$CURL_URL/$1" --head > /dev/null
 }
 
 remote_file_equals() {
 	local file="$1"
 	local remote="$2"
 	[ -z "$remote" ] && remote="$file"
-	curl -s "$CURL_URL/$remote" | diff - "$file" > /dev/null
+	$CURL -s "$CURL_URL/$remote" | diff - -- "$file" > /dev/null
 }
 
 assertContains() {
